import os
try:
    import cStringIO as StringIO
except ImportError:
    import StringIO

# Python 2.4 support: os lacks SEEK_END and friends
try:
    getattr(os, "SEEK_END")
except AttributeError:
    os.SEEK_SET, os.SEEK_CUR, os.SEEK_END = range(3)


class SimpleBuffer(object):
    """
    >>> b = SimpleBuffer()
    >>> b.write('abcdef')
    >>> b.read(3)
    'abc'
    >>> b.consume(3)
    >>> b.write('z')
    >>> b.read()
    'defz'
    >>> b.read()
    'defz'
    >>> b.read(0)
    ''
    >>> repr(b)
    "<SimpleBuffer of 4 bytes, 7 total size, 'defz'>"
    >>> str(b)
    "<SimpleBuffer of 4 bytes, 7 total size, 'defz'>"
    >>> len(b)
    4
    >>> bool(b)
    True
    >>> b.flush()
    >>> len(b)
    0
    >>> bool(b)
    False
    >>> b.read(1)
    ''
    >>> b.write('a'*524288)
    >>> b.flush() # run GC code
    """
    def __init__(self):
        self.buf = StringIO.StringIO()
        self.size = 0
        self.offset = 0

    def write(self, data):
        self.buf.write(data)
        self.size += len(data)

    def read(self, size=None):
        self.buf.seek(self.offset)

        if size is None:
            data = self.buf.read()
        else:
            data = self.buf.read(size)

        self.buf.seek(0, os.SEEK_END)
        return data

    def consume(self, size):
        self.offset += size
        self.size -= size
        # GC old StringIO instance and free memory used by it.
        if self.size == 0 and self.offset > 524288:
            self.buf.close()
            self.buf = StringIO.StringIO()
            self.offset = 0

    def flush(self):
        self.consume(self.size)


    def __nonzero__(self):
        return self.size > 0

    def __len__(self):
        return self.size

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<SimpleBuffer of %i bytes, %i total size, %r%s>' % \
                    (self.size, self.size + self.offset, self.read(16),
                     (self.size > 16) and '...' or '')



import select

def loop(clients):
    while True:
        for c in clients:
            c.run_any_callbacks()

        rfds = clients
        wfds = [c for c in clients if c.needs_write()]
        r, w, e = select.select(rfds, wfds, rfds)

        for c in r:
            c.on_read()
        for c in w:
            c.on_write()



import functools

from . import connection
from . import machine

def meta_attach_methods(name, bases, cls):
    decorator, list_of_methods = cls['attach_methods']
    for method in list_of_methods:
        cls[method.__name__] = decorator(method)
    return type(name, bases, cls)


def machine_decorator(method):
    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        callback = kwargs.get('callback')
        if callback is not None:
            del kwargs['callback']
        t = method(*args, **kwargs)
        t.user_callback = callback
        t.after_machine()
        return t.number
    return wrapper


class Client(connection.Connection):
    __metaclass__ = meta_attach_methods
    attach_methods = (machine_decorator, [
        machine.queue_declare,
        machine.queue_purge,
        machine.queue_delete,
        machine.basic_publish,
        machine.basic_consume,
        machine.basic_consume_multi,
        machine.basic_cancel,
        machine.basic_qos,
        machine.basic_get,
        machine.exchange_declare,
        machine.exchange_delete,
        machine.exchange_bind,
        machine.exchange_unbind,
        machine.queue_bind,
        machine.queue_unbind,
        ])

    @machine_decorator
    def connect(self):
        return self._connect()

    @machine_decorator
    def close(self):
        return self._close()

    def basic_ack(self, *args, **kwargs):
        machine.basic_ack(self, *args, **kwargs)

    def basic_reject(self, *args, **kwargs):
        machine.basic_reject(self, *args, **kwargs)




from client import Client

# Should be relative, but python 2.5 doesn't support importing start from that.
from spec_exceptions import *

from exceptions import ConnectionBroken

from poll import loop



import errno
import logging
import select
import socket
import struct
import time
import urllib
from . import urlparse

from . import channel
from . import exceptions
from . import machine
from . import simplebuffer
from . import spec
from . import promise

log = logging.getLogger('puka')



class Connection(object):
    frame_max = 131072

    def __init__(self, amqp_url='amqp:///', pubacks=None):
        self.pubacks = pubacks

        self.channels = channel.ChannelCollection()
        self.promises = promise.PromiseCollection(self)

        (self.username, self.password, self.vhost, self.host, self.port) = \
            parse_amqp_url(amqp_url)

    def _init_buffers(self):
        self.recv_buf = simplebuffer.SimpleBuffer()
        self.recv_need = 8
        self.send_buf = simplebuffer.SimpleBuffer()


    def fileno(self):
        return self.sd.fileno()

    def socket(self):
        return self.sd

    def _connect(self):
        self._handle_read = self._handle_conn_read
        self._init_buffers()

        try:
            addrinfo = socket.getaddrinfo(self.host, self.port, socket.AF_INET6, socket.SOCK_STREAM)
        except socket.gaierror:
            addrinfo = socket.getaddrinfo(self.host, self.port, socket.AF_INET, socket.SOCK_STREAM)

        (family, socktype, proto, canonname, sockaddr) = addrinfo[0]
        self.sd = socket.socket(family, socktype, proto)
        self.sd.setblocking(False)
        set_ridiculously_high_buffers(self.sd)
        try:
            self.sd.connect(sockaddr)
        except socket.error, e:
            if e.errno not in (errno.EINPROGRESS, errno.EWOULDBLOCK):
                raise
        return machine.connection_handshake(self)

    def on_read(self):
        try:
            r = self.sd.recv(131072)
        except socket.error, e:
            if e.errno == errno.EAGAIN:
                return
            else:
                raise

        if len(r) == 0:
            # a = self.sd.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
            self._shutdown(exceptions.mark_frame(spec.Frame(),
                                                 exceptions.ConnectionBroken()))

        self.recv_buf.write(r)

        if len(self.recv_buf) >= self.recv_need:
            data = self.recv_buf.read()
            offset = 0
            while len(data) - offset >= self.recv_need:
                offset, self.recv_need = \
                    self._handle_read(data, offset)
            self.recv_buf.consume(offset)

    def _handle_conn_read(self, data, offset):
        self._handle_read = self._handle_frame_read
        if data[offset:].startswith('AMQP'):
            a,b,c,d = struct.unpack('!BBBB', data[offset+4:offset+4+4])
            self._shutdown(exceptions.mark_frame(
                    spec.Frame(),
                    exceptions.UnsupportedProtocol("%s.%s.%s.%s" % (a,b,c,d))))
            return 0,0
        else:
            return self._handle_frame_read(data, offset)

    def _handle_frame_read(self, data, start_offset):
        offset = start_offset
        if len(data)-start_offset < 8:
            return start_offset, 8
        frame_type, channel, payload_size = \
            struct.unpack_from('!BHI', data, offset)
        offset += 7
        if len(data)-start_offset < 8+payload_size:
            return start_offset, 8+payload_size
        assert data[offset+payload_size] == '\xCE'

        if frame_type == 0x01: # Method frame
            method_id, = struct.unpack_from('!I', data, offset)
            offset += 4
            frame, offset = spec.METHODS[method_id](data, offset)
            self.channels.channels[channel].inbound_method(frame)
        elif frame_type == 0x02: # header frame
            (class_id, body_size) = struct.unpack_from('!HxxQ', data, offset)
            offset += 12
            props, offset = spec.PROPS[class_id](data, offset)
            self.channels.channels[channel].inbound_props(body_size, props)
        elif frame_type == 0x03: # body frame
            body_chunk = str(data[offset : offset+payload_size])
            self.channels.channels[channel].inbound_body(body_chunk)
            offset += len(body_chunk)
        else:
            assert False, "Unknown frame type"

        offset += 1 # '\xCE'
        assert offset == start_offset+8+payload_size
        return offset, 8


    def _send(self, data):
        p = bool(self.send_buf)
        self.send_buf.write(data)
        # Do not try to write straightaway, better wait for more data.

    def _send_frames(self, channel_number, frames):
        self._send( ''.join([''.join((struct.pack('!BHI',
                                                  frame_type,
                                                  channel_number,
                                                  len(payload)),
                                      payload, '\xCE')) \
                                 for frame_type, payload in frames]) )

    def needs_write(self):
        # Try to flush before answering.
        self.on_write()
        return bool(self.send_buf)

    def on_write(self):
        try:
            # On windows socket.send blows up if the buffer is too large.
            r = self.sd.send(self.send_buf.read(128*1024))
        except socket.error, e:
            if e.errno == errno.EAGAIN:
                return
            else:
                raise
        self.send_buf.consume(r)


    def _tune_frame_max(self, new_frame_max):
        new_frame_max = new_frame_max if new_frame_max != 0 else 2**19
        self.frame_max = min(self.frame_max, new_frame_max)
        return self.frame_max


    def wait(self, promise_numbers, timeout=None, raise_errors=True):
        '''
        Wait for selected promises. Exit after promise runs a callback.
        '''
        if timeout is not None:
            t1 = time.time() + timeout
        else:
            td = None

        if isinstance(promise_numbers, int):
            promise_numbers = [promise_numbers]
        promise_numbers = set(promise_numbers)
        while True:
            while True:
                ready = promise_numbers & self.promises.ready
                if not ready:
                    break
                promise_number = ready.pop()
                return self.promises.run_callback(promise_number,
                                                 raise_errors=raise_errors)

            if timeout is not None:
                t0 = time.time()
                td = t1 - t0
                if td < 0:
                    break

            r, w, e = select.select([self],
                                    [self] if self.needs_write() else [],
                                    [self],
                                    td)
            if r or e:
                self.on_read()
            if w:
                self.on_write()
            if not r and not e and not w:
                # timeout
                return None

    def wait_for_any(self):
        return self.loop()

    def wait_for_all(self, promise_list, raise_errors=True):
        for promise in promise_list:
            self.wait(promise, raise_errors=raise_errors)

    def loop(self, timeout=None):
        '''
        Wait for any promise. Block forever.
        '''
        if timeout is not None:
            t1 = time.time() + timeout
        else:
            td = None
        self._loop_break = False
        self.run_any_callbacks()
        while not self._loop_break:
            if timeout is not None:
                t0 = time.time()
                td = t1 - t0
                if td < 0:
                    break
            r, w, e = select.select([self],
                                    [self] if self.needs_write() else [],
                                    [self],
                                    td)
            if r or e:
                self.on_read()
            if w:
                self.on_write()
            self.run_any_callbacks()

    def loop_break(self):
        self._loop_break = True

    def run_any_callbacks(self):
        '''
        Run any callbacks, any promises, but do not block.
        '''
        while self.promises.ready:
            [self.promises.run_callback(promise, raise_errors=False) \
                 for promise in list(self.promises.ready)]


    def _shutdown(self, result):
        # Cancel all events.
        for promise in self.promises.all():
            # It's possible that a promise may be already `done` but still not
            # removed. For example due to `refcnt`. In that case don't run
            # callbacks.
            if promise.to_be_released is False:
                promise.done(result)

        # And kill the socket
        self.sd.shutdown(socket.SHUT_RDWR)
        self.sd.close()
        self.sd = None
        # Sending is illegal
        self.send_buf = None

    def _close(self):
        return machine.connection_close(self)

    def set_callback(self, promise_number, callback):
        promise = self.promises.by_number(promise_number)
        promise.user_callback = callback


def parse_amqp_url(amqp_url):
    '''
    >>> parse_amqp_url('amqp:///')
    ('guest', 'guest', '/', 'localhost', 5672)
    >>> parse_amqp_url('amqp://a:b@c:1/d')
    ('a', 'b', '/d', 'c', 1)
    >>> parse_amqp_url('amqp://g%20uest:g%20uest@host/vho%20st')
    ('g uest', 'g uest', '/vho st', 'host', 5672)
    >>> parse_amqp_url('http://asd')
    Traceback (most recent call last):
      ...
    AssertionError: Only amqp:// protocol supported.
    '''
    assert amqp_url.startswith('amqp://'), "Only amqp:// protocol supported."
    # urlsplit doesn't know how to parse query when scheme is amqp,
    # we need to pretend we're http'
    o = urlparse.urlsplit('http://' + amqp_url[len('amqp://'):])
    username = urllib.unquote(o.username) if o.username else 'guest'
    password = urllib.unquote(o.password) if o.password else 'guest'

    vhost = urllib.unquote(o.path) if o.path else '/'
    host = o.hostname or 'localhost'
    port = o.port if o.port else 5672
    return (username, password, vhost, host, port)

def set_ridiculously_high_buffers(sd):
    '''
    Set large tcp/ip buffers kernel. Let's move the complexity
    to the operating system! That's a wonderful idea!
    '''
    for flag in [socket.SO_SNDBUF, socket.SO_RCVBUF]:
        for i in range(10):
            bef = sd.getsockopt(socket.SOL_SOCKET, flag)
            try:
                sd.setsockopt(socket.SOL_SOCKET, flag, bef*2)
            except socket.error:
                break
            aft = sd.getsockopt(socket.SOL_SOCKET, flag)
            if aft <= bef or aft >= 1024*1024:
                break



import copy
import logging

from . import exceptions
from . import spec

log = logging.getLogger('puka')

def _nothing(t):
    pass

####
def connection_handshake(conn):
    # Bypass conn._send, we want the socket to be writable first.
    conn.send_buf.write(spec.PREAMBLE)
    t = conn.promises.new(_connection_handshake, reentrant=True)
    conn.x_connection_promise = t
    return t

def _connection_handshake(t):
    assert t.channel.number == 0
    t.register(spec.METHOD_CONNECTION_START, _connection_start)

def _connection_start(t, result):
    # log.info("Connected to %r", result['server_properties'])
    assert 'PLAIN' in result['mechanisms'].split(), "Only PLAIN auth supported."
    response = '\0%s\0%s' % (t.conn.username, t.conn.password)
    frames = spec.encode_connection_start_ok({'product': 'Puka'}, 'PLAIN',
                                             response, 'en_US')
    t.register(spec.METHOD_CONNECTION_TUNE, _connection_tune)
    t.send_frames(frames)
    t.x_cached_result = result
    t.conn.x_server_props = result['server_properties']
    try:
        t.conn.x_server_version = \
            map(int, t.conn.x_server_props['version'].split('.'))
    except ValueError:
        t.conn.x_server_version = (Ellipsis,)
    if t.conn.pubacks is None:
        t.conn.x_pubacks = t.conn.x_server_version > (2,3,0)
    else:
        t.conn.x_pubacks = t.conn.pubacks


def _connection_tune(t, result):
    frame_max = t.conn._tune_frame_max(result['frame_max'])
    channel_max = t.conn.channels.tune_channel_max(result['channel_max'])

    t.register(spec.METHOD_CONNECTION_OPEN_OK, _connection_open_ok)
    f1 = spec.encode_connection_tune_ok(channel_max, frame_max, 0)
    f2 = spec.encode_connection_open(t.conn.vhost)
    t.send_frames(f1 + f2)

def _connection_open_ok(ct, result):
    ct.register(spec.METHOD_CONNECTION_CLOSE, _connection_close)
    # Never free the promise and channel.
    ct.ping(ct.x_cached_result)
    ct.conn.x_connection_promise = ct
    publish_promise(ct.conn)


def publish_promise(conn):
    if conn.x_pubacks:
        pt = conn.promises.new(_pt_channel_open_ok_puback)
    else:
        pt = conn.promises.new(_pt_channel_open_ok)
    pt.x_async_enabled = False
    pt.x_delivery_tag = 1
    pt.x_delivery_tag_shift = 0
    pt.x_async_inflight = {}
    pt.x_async_next = []
    conn.x_publish_promise = pt

def _pt_channel_open_ok_puback(pt, _result=None):
    pt.send_frames( spec.encode_confirm_select() )
    pt.register(spec.METHOD_CONFIRM_SELECT_OK, _pt_channel_open_ok)
    pt.register(spec.METHOD_BASIC_ACK, _pt_basic_ack)

def _pt_channel_open_ok(pt, _result=None):
    pt.x_async_enabled = True
    pt.register(spec.METHOD_CHANNEL_CLOSE, _pt_channel_close)
    pt.register(spec.METHOD_BASIC_RETURN, _pt_basic_return)
    # Send remaining messages.
    _pt_async_flush(pt)

def fix_basic_publish_headers(headers):
    nheaders = {}
    nheaders.update(headers) # copy
    if 'persistent' in nheaders:
        if nheaders['persistent']:
            nheaders['delivery_mode'] = 2
        del nheaders['persistent']
    else:
        nheaders['delivery_mode'] = 2
    # That's not a good idea.
    assert 'headers' not in headers
    return nheaders

def basic_publish(conn, exchange, routing_key='', mandatory=False,
                  immediate=False, headers={}, body=''):
    pt = conn.x_publish_promise
    delivery_tag = pt.x_delivery_tag
    pt.x_delivery_tag += 1

    nheaders = fix_basic_publish_headers(headers)
    assert 'x-puka-delivery-tag' not in nheaders
    nheaders['x-puka-delivery-tag'] = delivery_tag

    frames = spec.encode_basic_publish(exchange, routing_key, mandatory,
                                       immediate, nheaders, body,
                                       conn.frame_max)
    if not conn.x_pubacks:
        # Construct ack packet.
        eheaders = {'x-puka-delivery-tag': delivery_tag, 'x-puka-footer': True}
        frames = frames + \
                 spec.encode_basic_publish('', '', False, True, eheaders,
                                                '', conn.frame_max)
    t = conn.promises.new(_nothing, no_channel=True)
    pt.x_async_next.append( (delivery_tag, t, frames) )
    _pt_async_flush(pt)
    return t

def _pt_async_flush(pt):
    if pt.x_async_enabled:
        frames_acc = []
        for delivery_tag, t, frames in pt.x_async_next:
            pt.x_async_inflight[delivery_tag] = t
            frames_acc.extend( frames )
        pt.x_async_next = []
        pt.send_frames(frames_acc)

def _pt_basic_return(pt, result):
    pt.register(spec.METHOD_BASIC_RETURN, _pt_basic_return)
    delivery_tag = result['headers']['x-puka-delivery-tag']
    if delivery_tag in pt.x_async_inflight:
        t = pt.x_async_inflight[delivery_tag]
        del pt.x_async_inflight[delivery_tag]
        if 'x-puka-footer' in result['headers']: # ok
            t.done(spec.Frame())
        else: # return
            exceptions.mark_frame(result)
            t.done(result)

def _pt_basic_ack(pt, result):
    pt.register(spec.METHOD_BASIC_ACK, _pt_basic_ack)
    delivery_tag = result['delivery_tag'] + pt.x_delivery_tag_shift
    if delivery_tag in pt.x_async_inflight:
        if result['multiple'] == True:
            delivery_tags = filter(lambda dt: dt <= delivery_tag,
                                   pt.x_async_inflight.iterkeys())
        else:
            delivery_tags = (delivery_tag,)
        for delivery_tag in delivery_tags:
            t = pt.x_async_inflight[delivery_tag]
            del pt.x_async_inflight[delivery_tag]
            t.done(spec.Frame())

def _pt_channel_close(pt, result):
    pt.x_async_enabled = False
    pt.x_delivery_tag_shift = pt.x_delivery_tag
    # Start off with reestablishing the channel
    if pt.conn.x_pubacks:
        pt.x_delivery_tag_shift -= 1 # starting from 1.
        pt.register(spec.METHOD_CHANNEL_OPEN_OK, _pt_channel_open_ok_puback)
    else:
        pt.register(spec.METHOD_CHANNEL_OPEN_OK, _pt_channel_open_ok)
    pt.send_frames( spec.encode_channel_close_ok() +
                    spec.encode_channel_open(''))
    # All the publishes are marked as failed.
    exceptions.mark_frame(result)
    for t in pt.x_async_inflight.itervalues():
        t.done(result)
    pt.x_async_inflight.clear()


def _connection_close(t, result):
    exceptions.mark_frame(result)
    t.ping(result)
    # Explode, kill everything.
    log.error('Connection killed with %r', result)
    t.conn._shutdown(result)

def connection_close(conn):
    t = conn.x_connection_promise
    t.register(spec.METHOD_CONNECTION_CLOSE_OK, _connection_close_ok)
    t.send_frames(spec.encode_connection_close(200, '', 0, 0))
    return t

def _connection_close_ok(t, result):
    # Ping this promise with success.
    t.ping(copy.copy(result))
    # Cancel all our promises with failure.
    exceptions.mark_frame(result)
    t.conn._shutdown(result)


####
def channel_open(t, callback):
    t.register(spec.METHOD_CHANNEL_OPEN_OK, _channel_open_ok)
    t.x_callback = callback
    t.send_frames( spec.encode_channel_open('') )

def _channel_open_ok(t, result):
    t.x_callback()


####
def queue_declare(conn, queue='', durable=False, exclusive=False,
                  auto_delete=False, arguments={}):
    t = conn.promises.new(_queue_declare)
    t.x_frames = spec.encode_queue_declare(queue, False, durable, exclusive,
                                           auto_delete, arguments)
    return t

def _queue_declare(t, result=None):
    t.register(spec.METHOD_QUEUE_DECLARE_OK, _queue_declare_ok)
    t.send_frames(t.x_frames)

def _queue_declare_ok(t, result):
    t.done(result)


####
def basic_consume(conn, queue, prefetch_count=0, no_local=False, no_ack=False,
                  exclusive=False, arguments={}):
    q = {'queue': queue,
         'no_local': no_local,
         'exclusive': exclusive,
         'arguments': arguments,
         }
    return basic_consume_multi(conn, [q], prefetch_count, no_ack)

####
def basic_consume_multi(conn, queues, prefetch_count=0, no_ack=False):
    t = conn.promises.new(_bcm_basic_qos, reentrant=True)
    t.x_frames = spec.encode_basic_qos(0, prefetch_count, False)
    t.x_consumes = []
    for item in queues:
        if isinstance(item, str):
            queue = item
            no_local = exclusive = False
            arguments = {}
        else:
            queue = item['queue']
            no_local = item.get('no_local', False)
            exclusive = item.get('exclusive', False)
            arguments = item.get('arguments', {})
        t.x_consumes.append( (queue, spec.encode_basic_consume(
                    queue, '', no_local, no_ack, exclusive, arguments)) )
    t.x_no_ack = no_ack
    t.x_consumer_tag = {}
    t.register(spec.METHOD_BASIC_DELIVER, _bcm_basic_deliver)
    return t

def _bcm_basic_qos(t):
    t.register(spec.METHOD_BASIC_QOS_OK, _bcm_basic_qos_ok)
    t.send_frames(t.x_frames)

def _bcm_basic_qos_ok(t, result):
    _bcm_send_basic_consume(t)

def _bcm_send_basic_consume(t):
    t.register(spec.METHOD_BASIC_CONSUME_OK, _bcm_basic_consume_ok)
    t.x_queue, frames = t.x_consumes.pop()
    t.send_frames(frames)

def _bcm_basic_consume_ok(t, consume_result):
    t.x_consumer_tag[t.x_queue] = consume_result['consumer_tag']
    if t.x_consumes:
        _bcm_send_basic_consume(t)

def _bcm_basic_deliver(t, msg_result):
    t.register(spec.METHOD_BASIC_DELIVER, _bcm_basic_deliver)
    msg_result['promise_number'] = t.number
    if t.x_no_ack is False:
        t.refcnt_inc()
    t.ping(msg_result)


##
def basic_ack(conn, msg_result):
    t = conn.promises.by_number(msg_result['promise_number'])
    t.send_frames( spec.encode_basic_ack(msg_result['delivery_tag'], False) )
    assert t.x_no_ack is False
    t.refcnt_dec()
    return t

##
def basic_reject(conn, msg_result):
    t = conn.promises.by_number(msg_result['promise_number'])
    # For basic.reject requeue must be True.
    t.send_frames( spec.encode_basic_reject(msg_result['delivery_tag'], True) )
    assert t.x_no_ack is False
    t.refcnt_dec()
    return t

##
def basic_qos(conn, consume_promise, prefetch_count=0):
    # TODO: race?
    t = conn.promises.new(_basic_qos, no_channel=True)
    t.x_ct = conn.promises.by_number(consume_promise)
    t.x_frames = spec.encode_basic_qos(0, prefetch_count, False)
    return t

def _basic_qos(t):
    ct = t.x_ct
    ct.register(spec.METHOD_BASIC_QOS_OK, _basic_qos_ok)
    ct.send_frames( t.x_frames )
    ct.x_qos_promise = t

def _basic_qos_ok(ct, result):
    t = ct.x_qos_promise
    t.done(result)

##
def basic_cancel(conn, consume_promise):
    # TODO: race?
    t = conn.promises.new(_basic_cancel, no_channel=True)
    t.x_ct = conn.promises.by_number(consume_promise)
    return t

def _basic_cancel(t):
    t.x_ct.x_mt = t
    _basic_cancel_one(t.x_ct)

def _basic_cancel_one(ct):
    consumer_tag = ct.x_consumer_tag.pop(ct.x_consumer_tag.keys()[0])
    ct.register(spec.METHOD_BASIC_CANCEL_OK, _basic_cancel_ok)
    ct.send_frames( spec.encode_basic_cancel(consumer_tag) )

def _basic_cancel_ok(ct, result):
    if ct.x_consumer_tag:
        _basic_cancel_one(ct)
    else:
        ct.x_mt.done(result)
        ct.x_mt = None
        ct.done(None, no_callback=True)
        ct.refcnt_clear()

####
def basic_get(conn, queue, no_ack=False):
    t = conn.promises.new(_basic_get)
    t.x_frames = spec.encode_basic_get(queue, no_ack)
    t.x_no_ack = no_ack
    return t

def _basic_get(t):
    t.register(spec.METHOD_BASIC_GET_OK, _basic_get_ok)
    t.register(spec.METHOD_BASIC_GET_EMPTY, _basic_get_empty)
    t.send_frames(t.x_frames)

def _basic_get_ok(t, msg_result):
    msg_result['promise_number'] = t.number
    if t.x_no_ack is False:
        t.refcnt_inc()
    t.done(msg_result)

def _basic_get_empty(t, result):
    result['empty'] = True
    t.done(result)


####
def exchange_declare(conn, exchange, type='direct', durable=False,
                     auto_delete=False, arguments={}):
    t = conn.promises.new(_exchange_declare)

    t.x_frames = spec.encode_exchange_declare(exchange, type, False, durable,
                                              auto_delete, False, arguments)
    return t

def _exchange_declare(t, result=None):
    t.register(spec.METHOD_EXCHANGE_DECLARE_OK, _exchange_declare_ok)
    t.send_frames(t.x_frames)

def _exchange_declare_ok(t, result):
    t.done(result)


####
def _generic_callback(t):
    t.register(t.x_method, _generic_callback_ok)
    t.send_frames(t.x_frames)

def _generic_callback_ok(t, result):
    t.done(result)

####
def exchange_delete(conn, exchange, if_unused=False):
    t = conn.promises.new(_generic_callback)
    t.x_method = spec.METHOD_EXCHANGE_DELETE_OK
    t.x_frames = spec.encode_exchange_delete(exchange, if_unused)
    return t

def exchange_bind(conn, destination, source, routing_key='', arguments={}):
    t = conn.promises.new(_generic_callback)
    t.x_method = spec.METHOD_EXCHANGE_BIND_OK
    t.x_frames = spec.encode_exchange_bind(destination, source, routing_key,
                                           arguments)
    return t

def exchange_unbind(conn, destination, source, routing_key='', arguments={}):
    t = conn.promises.new(_generic_callback)
    t.x_method = spec.METHOD_EXCHANGE_UNBIND_OK
    t.x_frames = spec.encode_exchange_unbind(destination, source, routing_key,
                                             arguments)
    return t

def queue_delete(conn, queue, if_unused=False, if_empty=False):
    t = conn.promises.new(_generic_callback)
    t.x_method = spec.METHOD_QUEUE_DELETE_OK
    t.x_frames = spec.encode_queue_delete(queue, if_unused, if_empty)
    return t

def queue_purge(conn, queue):
    t = conn.promises.new(_generic_callback)
    t.x_method = spec.METHOD_QUEUE_PURGE_OK
    t.x_frames = spec.encode_queue_purge(queue)
    return t

def queue_bind(conn, queue, exchange, routing_key='', arguments={}):
    t = conn.promises.new(_generic_callback)
    t.x_method = spec.METHOD_QUEUE_BIND_OK
    t.x_frames = spec.encode_queue_bind(queue, exchange, routing_key,
                                        arguments)
    return t

def queue_unbind(conn, queue, exchange, routing_key='', arguments={}):
    t = conn.promises.new(_generic_callback)
    t.x_method = spec.METHOD_QUEUE_UNBIND_OK
    t.x_frames = spec.encode_queue_unbind(queue, exchange, routing_key,
                                          arguments)
    return t




import logging

from . import channel
from . import spec
from . import exceptions

log = logging.getLogger('puka')



class PromiseCollection(object):
    def __init__(self, conn):
        self.conn = conn
        self._promises = {}
        self.promise_number = 1
        self.ready = set()

    def new(self, on_channel, **kwargs):
        number = self.promise_number
        self.promise_number += 1
        promise = Promise(self.conn, number, on_channel, **kwargs)
        self._promises[number] = promise
        return promise

    def free(self, promise):
        del self._promises[promise.number]

    def mark_ready(self, promise):
        self.ready.add( promise.number )

    def unmark_ready(self, promise):
        self.ready.remove( promise.number )

    def run_callback(self, number, **kwargs):
        return self._promises[number].run_callback(**kwargs)

    def by_number(self, number):
        return self._promises[number]

    def all(self):
        return self._promises.values()


class Promise(object):
    to_be_released = False
    delay_release = None
    user_callback = None
    after_machine_callback = None
    refcnt = 0

    def __init__(self, conn, number, on_channel, reentrant=False,
                 no_channel=False):
        self.number = number
        self.conn = conn
        self.on_channel = on_channel
        self.reentrant = reentrant

        self.methods = {}
        self.callbacks = []

        if not no_channel:
            self.conn.channels.allocate(self, self._on_channel)
        else:
            self.channel = None
            self.after_machine_callback = self._on_channel

    def restore_error_handler(self):
        self.register(spec.METHOD_CHANNEL_CLOSE, self._on_channel_close)

    def _on_channel(self):
        if self.channel:
            self.channel.alive = True
        self.restore_error_handler()
        self.on_channel(self)

    def _on_channel_close(self, _t, result):
        # log.warn('channel %i died %r', self.channel.number, result)
        exceptions.mark_frame(result)
        self.send_frames(spec.encode_channel_close_ok())
        if self.channel:
            self.channel.alive = False
        self.done(result)

    def recv_method(self, result):
        # log.debug('#%i recv_method %r', self.number, result)
        # In this order, to allow callback to re-register to the same method.
        callback = self.methods[result.method_id]
        del self.methods[result.method_id]
        callback(self, result)

    def register(self, method_id, callback):
        self.methods[method_id] = callback

    def unregister(self, method_id):
        del self.methods[method_id]


    def send_frames(self, frames):
        self.conn._send_frames(self.channel.number, frames)


    def done(self, result, delay_release=None, no_callback=False):
        # log.debug('#%i done %r', self.number, result)
        assert self.to_be_released == False
        if not self.reentrant:
            assert len(self.callbacks) == 0
        if not no_callback:
            self.callbacks.append( (self.user_callback, result) )
        else:
            self.callbacks.append( (None, result) )
        self.conn.promises.mark_ready(self)
        self.to_be_released = True
        self.delay_release = delay_release
        self.methods.clear()
        self.restore_error_handler()

    def ping(self, result):
        assert self.to_be_released == False
        assert self.reentrant
        self.callbacks.append( (self.user_callback, result) )
        self.conn.promises.mark_ready(self)


    def run_callback(self, raise_errors=True):
        user_callback, result = self.callbacks.pop(0)
        if not self.callbacks:
            self.conn.promises.unmark_ready(self)
        if user_callback:
            user_callback(self.number, result)
            # We may be already freed now.
            # (consider basic_get + basic_ack inside callback)

        self.maybe_release()
        if raise_errors and result.is_error:
            raise result.exception
        return result


    def after_machine(self):
        if self.after_machine_callback:
            self.after_machine_callback()
            self.after_machine_callback = None

    def refcnt_inc(self):
        self.refcnt += 1

    def refcnt_dec(self):
        assert self.refcnt > 0
        self.refcnt -= 1
        if self.refcnt == 0:
            self.maybe_release()

    def refcnt_clear(self):
        self.refcnt = 0

    def maybe_release(self):
        # If not released yet, not used by callbacks, and not refcounted.
        if ( self.number is not None and not self.callbacks and
             self.to_be_released and self.refcnt == 0):
            self._release()

    def _release(self):
        # Release channel and unlink self.
        if self.delay_release is None:
            if self.channel:
                self.conn.channels.deallocate(self.channel)
            self.conn.promises.free(self)
        elif self.delay_release is Ellipsis:
            # Never free.
            pass
        else:
            # TODO:
            print "Unable to free channel %i (promise %i)" % \
                (self.channel.number, self.number)
        self.number = None



import socket
from . import spec_exceptions

class ConnectionBroken(socket.error): pass
class UnsupportedProtocol(socket.error): pass


def exception_from_frame(result):
    reply_code = result.get('reply_code', 0)
    if reply_code in spec_exceptions.ERRORS:
        return spec_exceptions.ERRORS[reply_code](result)
    return spec_exceptions.AMQPError(result)

def mark_frame(result, exception=None):
    result.is_error = True
    if exception is None:
        result.exception = exception_from_frame(result)
    else:
        result.exception = exception
    return result



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0
#
# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
# the License for the specific language governing rights and
# limitations under the License.
#
# The Original Code is Pika.
#
# The Initial Developers of the Original Code are LShift Ltd, Cohesive
# Financial Technologies LLC, and Rabbit Technologies Ltd.  Portions
# created before 22-Nov-2008 00:00:00 GMT by LShift Ltd, Cohesive
# Financial Technologies LLC, or Rabbit Technologies Ltd are Copyright
# (C) 2007-2008 LShift Ltd, Cohesive Financial Technologies LLC, and
# Rabbit Technologies Ltd.
#
# Portions created by LShift Ltd are Copyright (C) 2007-2009 LShift
# Ltd. Portions created by Cohesive Financial Technologies LLC are
# Copyright (C) 2007-2009 Cohesive Financial Technologies
# LLC. Portions created by Rabbit Technologies Ltd are Copyright (C)
# 2007-2009 Rabbit Technologies Ltd.
#
# Portions created by Tony Garnock-Jones are Copyright (C) 2009-2010
# LShift Ltd and Tony Garnock-Jones.
#
# All Rights Reserved.
#
# Contributor(s): ______________________________________.
#
# Alternatively, the contents of this file may be used under the terms
# of the GNU General Public License Version 2 or later (the "GPL"), in
# which case the provisions of the GPL are applicable instead of those
# above. If you wish to allow use of your version of this file only
# under the terms of the GPL, and not to allow others to use your
# version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the
# notice and other provisions required by the GPL. If you do not
# delete the provisions above, a recipient may use your version of
# this file under the terms of any one of the MPL or the GPL.
#
# ***** END LICENSE BLOCK *****
#
# Code originally from Pika, adapted to Puka.
#

import struct
import decimal
import datetime
import calendar


def encode(table):
    r'''
    >>> encode(None)
    '\x00\x00\x00\x00'
    >>> encode({})
    '\x00\x00\x00\x00'
    >>> encode({'a':1, 'c':1, 'd':'x', 'e':{}})
    '\x00\x00\x00\x1d\x01aI\x00\x00\x00\x01\x01cI\x00\x00\x00\x01\x01eF\x00\x00\x00\x00\x01dS\x00\x00\x00\x01x'
    >>> encode({'a':decimal.Decimal('1.0')})
    '\x00\x00\x00\x08\x01aD\x00\x00\x00\x00\x01'
    >>> encode({'a':decimal.Decimal('5E-3')})
    '\x00\x00\x00\x08\x01aD\x03\x00\x00\x00\x05'
    >>> encode({'a':datetime.datetime(2010,12,31,23,58,59)})
    '\x00\x00\x00\x0b\x01aT\x00\x00\x00\x00M\x1enC'
    >>> encode({'test':decimal.Decimal('-0.01')})
    '\x00\x00\x00\x0b\x04testD\x02\xff\xff\xff\xff'
    >>> encode({'a':-1, 'b':[1,2,3,4,-1],'g':-1})
    '\x00\x00\x00.\x01aI\xff\xff\xff\xff\x01bA\x00\x00\x00\x19I\x00\x00\x00\x01I\x00\x00\x00\x02I\x00\x00\x00\x03I\x00\x00\x00\x04I\xff\xff\xff\xff\x01gI\xff\xff\xff\xff'
    >>> encode({'a': True, 'b':False})
    '\x00\x00\x00\x08\x01at\x01\x01bt\x00'
    '''
    pieces = []
    if table is None:
        table = {}
    length_index = len(pieces)
    pieces.append(None) # placeholder
    tablesize = 0
    for (key, value) in table.iteritems():
        pieces.append(struct.pack('B', len(key)))
        pieces.append(key)
        tablesize = tablesize + 1 + len(key)
        tablesize += encode_value(pieces, value)
    pieces[length_index] = struct.pack('>I', tablesize)
    return ''.join(pieces)

def encode_value(pieces, value):
    if isinstance(value, str):
        pieces.append(struct.pack('>cI', 'S', len(value)))
        pieces.append(value)
        return 5 + len(value)
    elif isinstance(value, bool):
        pieces.append(struct.pack('>cB', 't', int(value)))
        return 2
    elif isinstance(value, int):
        pieces.append(struct.pack('>ci', 'I', value))
        return 5
    elif isinstance(value, long):
        pieces.append(struct.pack('>cq', 'l', value))
        return 9
    elif isinstance(value, decimal.Decimal):
        value = value.normalize()
        if value._exp < 0:
            decimals = -value._exp
            raw = int(value * (decimal.Decimal(10) ** decimals))
            pieces.append(struct.pack('>cBi', 'D', decimals, raw))
        else:
            # per spec, the "decimals" octet is unsigned (!)
            pieces.append(struct.pack('>cBi', 'D', 0, int(value)))
        return 6
    elif isinstance(value, datetime.datetime):
        pieces.append(struct.pack('>cQ', 'T', calendar.timegm(
                    value.utctimetuple())))
        return 9
    elif isinstance(value, dict):
        pieces.append(struct.pack('>c', 'F'))
        piece = encode(value)
        pieces.append(piece)
        return 1 + len(piece)
    elif isinstance(value, list):
        p = []
        [encode_value(p, v) for v in value]
        piece = ''.join(p)
        pieces.append(struct.pack('>cI', 'A', len(piece)))
        pieces.append(piece)
        return 5 + len(piece)
    else:
        assert False, "Unsupported field kind during encoding %r/%r" % (
            key, value)

def decode(encoded, offset):
    '''
    >>> decode(encode(None), 0)
    ({}, 4)
    >>> decode(encode({}), 0)[0]
    {}
    >>> decode(encode({'a':1, 'c':1, 'd':'x', 'e':{}, 'f':-1}), 0)[0]
    {'a': 1, 'c': 1, 'e': {}, 'd': 'x', 'f': -1}

    # python 2.5 reports Decimal("1.01"), python 2.6 Decimal('1.01')
    >>> decode(encode({'a':decimal.Decimal('1.01')}), 0)[0] # doctest: +ELLIPSIS
    {'a': Decimal(...1.01...)}
    >>> decode(encode({'a':decimal.Decimal('5E-30')}), 0)[0] # doctest: +ELLIPSIS
    {'a': Decimal(...5E-30...)}
    >>> decode(encode({'a':datetime.datetime(2010,12,31,23,58,59)}), 0)[0]
    {'a': datetime.datetime(2010, 12, 31, 23, 58, 59)}
    >>> decode(encode({'test':decimal.Decimal('-1.234')}), 0)[0]
    {'test': Decimal('-1.234')}
    >>> decode(encode({'test':decimal.Decimal('1.234')}), 0)[0]
    {'test': Decimal('1.234')}
    >>> decode(encode({'test':decimal.Decimal('1000000')}), 0)[0]
    {'test': Decimal('1000000')}
    >>> decode(encode({'a':[1,2,3,'a',decimal.Decimal('-0.01')]}), 0)[0]
    {'a': [1, 2, 3, 'a', Decimal('-0.01')]}
    >>> decode(encode({'a': 0x7EADBEEFDEADBEEFL}), 0)[0]
    {'a': 9128161957192253167L}
    >>> decode(encode({'a': True, 'b': False}), 0)[0]
    {'a': True, 'b': False}
    '''
    result = {}
    tablesize = struct.unpack_from('>I', encoded, offset)[0]
    offset = offset + 4
    limit = offset + tablesize
    while offset < limit:
        keylen = struct.unpack_from('B', encoded, offset)[0]
        offset = offset + 1
        key = encoded[offset : offset + keylen]
        offset = offset + keylen
        result[key], offset = decode_value(encoded, offset)
    return (result, offset)

def decode_value(encoded, offset):
    kind = encoded[offset]
    offset = offset + 1
    if kind == 'S':
        length = struct.unpack_from('>I', encoded, offset)[0]
        offset = offset + 4
        value = encoded[offset : offset + length]
        offset = offset + length
    elif kind == 't':
        value = struct.unpack_from('>B', encoded, offset)[0]
        value = bool(value)
        offset = offset + 1
    elif kind == 'I':
        value = struct.unpack_from('>i', encoded, offset)[0]
        offset = offset + 4
    elif kind == 'l':
        value = struct.unpack_from('>q', encoded, offset)[0]
        value = long(value)
        offset = offset + 8
    elif kind == 'D':
        decimals = struct.unpack_from('B', encoded, offset)[0]
        offset = offset + 1
        raw = struct.unpack_from('>i', encoded, offset)[0]
        offset = offset + 4
        value = decimal.Decimal(raw) * (decimal.Decimal(10) ** -decimals)
    elif kind == 'T':
        value = datetime.datetime.utcfromtimestamp(
            struct.unpack_from('>Q', encoded, offset)[0])
        offset = offset + 8
    elif kind == 'F':
        (value, offset) = decode(encoded, offset)
    elif kind == 'A':
        length, = struct.unpack_from('>I', encoded, offset)
        offset = offset + 4
        offset_end = offset + length
        value = []
        while offset < offset_end:
            v, offset = decode_value(encoded, offset)
            value.append(v)
        assert offset == offset_end
    else:
        assert False, "Unsupported field kind %s during decoding" % (kind,)
    return value, offset



import array
import logging

from . import machine

log = logging.getLogger('puka')



class ChannelCollection(object):
    channel_max = 65535

    def __init__(self):
        self.channels = {}
        self.free_channels = []
        # Channel 0 is a special case.
        self.free_channel_numbers = [0]
        zero_channel = self.new()
        self.free_channels.append( zero_channel )

    def tune_channel_max(self, new_channel_max):
        new_channel_max = new_channel_max if new_channel_max != 0 else 65535
        self.channel_max = min(self.channel_max, new_channel_max)
        self.free_channel_numbers = array.array('H',
                                                xrange(self.channel_max, 0, -1))
        return self.channel_max

    def new(self):
        # TODO: handle out of channels case
        number = self.free_channel_numbers.pop()
        channel = Channel(number)
        self.channels[number] = channel
        return channel

    def allocate(self, promise, on_channel):
        if self.free_channels:
            channel = self.free_channels.pop()
            channel.promise = promise
            promise.channel = channel
            promise.after_machine_callback = on_channel
        else:
            channel = self.new()
            channel.promise = promise
            promise.channel = channel
            machine.channel_open(promise, on_channel)
        return channel

    def deallocate(self, channel):
        channel.promise.channel = channel.promise = None
        if channel.alive:
            self.free_channels.append( channel )
        else:
            del self.channels[channel.number]
            self.free_channel_numbers.append( channel.number )



class Channel(object):
    alive = False

    def __init__(self, number):
        self.number = number
        self.promise = None
        self._clear_inbound_state()

    def _clear_inbound_state(self):
        self.method_frame = self.props = None
        self.body_chunks = []
        self.body_len = self.body_size = 0


    def inbound_method(self, frame):
        if frame.has_content:
            self.method_frame = frame
        else:
            self._handle_inbound(frame)

    def inbound_props(self, body_size, props):
        self.body_size = body_size
        self.props = props
        if self.body_size == 0: # don't expect body frame
            self.inbound_body('')

    def inbound_body(self, body_chunk):
        self.body_chunks.append( body_chunk )
        self.body_len += len(body_chunk)
        if self.body_len == self.body_size:
            result = self.method_frame
            props = self.props

            result['body'] = ''.join(self.body_chunks)
            result['headers'] = props.get('headers', {})
            # Aint need a reference loop.
            if 'headers' in props:
                del props['headers']
            result['headers'].update( props )
            # Fix delivery_mode.
            result['headers']['persistent'] = \
                result['headers'].get('delivery_mode', 1) == 2

            self._clear_inbound_state()
            return self._handle_inbound(result)

    def _handle_inbound(self, result):
        self.promise.recv_method(result)




# Autogenerated - do not edit
import calendar
import datetime
import struct

from . import table


PREAMBLE = 'AMQP\x00\x00\x09\x01'

METHOD_CONNECTION_START         = 0x000A000A 	# 10,10 655370
METHOD_CONNECTION_START_OK      = 0x000A000B 	# 10,11 655371
METHOD_CONNECTION_SECURE        = 0x000A0014 	# 10,20 655380
METHOD_CONNECTION_SECURE_OK     = 0x000A0015 	# 10,21 655381
METHOD_CONNECTION_TUNE          = 0x000A001E 	# 10,30 655390
METHOD_CONNECTION_TUNE_OK       = 0x000A001F 	# 10,31 655391
METHOD_CONNECTION_OPEN          = 0x000A0028 	# 10,40 655400
METHOD_CONNECTION_OPEN_OK       = 0x000A0029 	# 10,41 655401
METHOD_CONNECTION_CLOSE         = 0x000A0032 	# 10,50 655410
METHOD_CONNECTION_CLOSE_OK      = 0x000A0033 	# 10,51 655411
METHOD_CHANNEL_OPEN             = 0x0014000A 	# 20,10 1310730
METHOD_CHANNEL_OPEN_OK          = 0x0014000B 	# 20,11 1310731
METHOD_CHANNEL_FLOW             = 0x00140014 	# 20,20 1310740
METHOD_CHANNEL_FLOW_OK          = 0x00140015 	# 20,21 1310741
METHOD_CHANNEL_CLOSE            = 0x00140028 	# 20,40 1310760
METHOD_CHANNEL_CLOSE_OK         = 0x00140029 	# 20,41 1310761
METHOD_EXCHANGE_DECLARE         = 0x0028000A 	# 40,10 2621450
METHOD_EXCHANGE_DECLARE_OK      = 0x0028000B 	# 40,11 2621451
METHOD_EXCHANGE_DELETE          = 0x00280014 	# 40,20 2621460
METHOD_EXCHANGE_DELETE_OK       = 0x00280015 	# 40,21 2621461
METHOD_EXCHANGE_BIND            = 0x0028001E 	# 40,30 2621470
METHOD_EXCHANGE_BIND_OK         = 0x0028001F 	# 40,31 2621471
METHOD_EXCHANGE_UNBIND          = 0x00280028 	# 40,40 2621480
METHOD_EXCHANGE_UNBIND_OK       = 0x00280033 	# 40,51 2621491
METHOD_QUEUE_DECLARE            = 0x0032000A 	# 50,10 3276810
METHOD_QUEUE_DECLARE_OK         = 0x0032000B 	# 50,11 3276811
METHOD_QUEUE_BIND               = 0x00320014 	# 50,20 3276820
METHOD_QUEUE_BIND_OK            = 0x00320015 	# 50,21 3276821
METHOD_QUEUE_PURGE              = 0x0032001E 	# 50,30 3276830
METHOD_QUEUE_PURGE_OK           = 0x0032001F 	# 50,31 3276831
METHOD_QUEUE_DELETE             = 0x00320028 	# 50,40 3276840
METHOD_QUEUE_DELETE_OK          = 0x00320029 	# 50,41 3276841
METHOD_QUEUE_UNBIND             = 0x00320032 	# 50,50 3276850
METHOD_QUEUE_UNBIND_OK          = 0x00320033 	# 50,51 3276851
METHOD_BASIC_QOS                = 0x003C000A 	# 60,10 3932170
METHOD_BASIC_QOS_OK             = 0x003C000B 	# 60,11 3932171
METHOD_BASIC_CONSUME            = 0x003C0014 	# 60,20 3932180
METHOD_BASIC_CONSUME_OK         = 0x003C0015 	# 60,21 3932181
METHOD_BASIC_CANCEL             = 0x003C001E 	# 60,30 3932190
METHOD_BASIC_CANCEL_OK          = 0x003C001F 	# 60,31 3932191
METHOD_BASIC_PUBLISH            = 0x003C0028 	# 60,40 3932200
METHOD_BASIC_RETURN             = 0x003C0032 	# 60,50 3932210
METHOD_BASIC_DELIVER            = 0x003C003C 	# 60,60 3932220
METHOD_BASIC_GET                = 0x003C0046 	# 60,70 3932230
METHOD_BASIC_GET_OK             = 0x003C0047 	# 60,71 3932231
METHOD_BASIC_GET_EMPTY          = 0x003C0048 	# 60,72 3932232
METHOD_BASIC_ACK                = 0x003C0050 	# 60,80 3932240
METHOD_BASIC_REJECT             = 0x003C005A 	# 60,90 3932250
METHOD_BASIC_RECOVER_ASYNC      = 0x003C0064 	# 60,100 3932260
METHOD_BASIC_RECOVER            = 0x003C006E 	# 60,110 3932270
METHOD_BASIC_RECOVER_OK         = 0x003C006F 	# 60,111 3932271
METHOD_BASIC_NACK               = 0x003C0078 	# 60,120 3932280
METHOD_CONFIRM_SELECT           = 0x0055000A 	# 85,10 5570570
METHOD_CONFIRM_SELECT_OK        = 0x0055000B 	# 85,11 5570571

CLASS_BASIC             = 0x003C




class Frame(dict):
    has_content = False
    is_error = False


class FrameConnectionStart(Frame):
    name = 'connection.start'
    method_id = METHOD_CONNECTION_START

def decode_connection_start(data, offset):
    frame = FrameConnectionStart()
    (frame['version_major'],
     frame['version_minor']) = struct.unpack_from('!BB', data, offset)
    offset += 1+1
    frame['server_properties'], offset = table.decode(data, offset)
    (str_len,) = struct.unpack_from('!I', data, offset)
    offset += 4
    frame['mechanisms'] = data[offset : offset+str_len]
    offset += str_len
    (str_len,) = struct.unpack_from('!I', data, offset)
    offset += 4
    frame['locales'] = data[offset : offset+str_len]
    offset += str_len
    return frame, offset


class FrameConnectionSecure(Frame):
    name = 'connection.secure'
    method_id = METHOD_CONNECTION_SECURE

def decode_connection_secure(data, offset):
    frame = FrameConnectionSecure()
    (str_len,) = struct.unpack_from('!I', data, offset)
    offset += 4
    frame['challenge'] = data[offset : offset+str_len]
    offset += str_len
    return frame, offset


class FrameConnectionTune(Frame):
    name = 'connection.tune'
    method_id = METHOD_CONNECTION_TUNE

def decode_connection_tune(data, offset):
    frame = FrameConnectionTune()
    (frame['channel_max'],
     frame['frame_max'],
     frame['heartbeat']) = struct.unpack_from('!HIH', data, offset)
    offset += 2+4+2
    return frame, offset


class FrameConnectionOpenOk(Frame):
    name = 'connection.open_ok'
    method_id = METHOD_CONNECTION_OPEN_OK

def decode_connection_open_ok(data, offset):
    frame = FrameConnectionOpenOk()
    (str_len,) = struct.unpack_from('!B', data, offset)
    offset += 1
    frame['known_hosts'] = data[offset : offset+str_len]
    offset += str_len
    return frame, offset


class FrameConnectionClose(Frame):
    name = 'connection.close'
    method_id = METHOD_CONNECTION_CLOSE

def decode_connection_close(data, offset):
    frame = FrameConnectionClose()
    (frame['reply_code'],
     str_len) = struct.unpack_from('!HB', data, offset)
    offset += 2+1
    frame['reply_text'] = data[offset : offset+str_len]
    offset += str_len
    (frame['class_id'],
     frame['method_id']) = struct.unpack_from('!HH', data, offset)
    offset += 2+2
    return frame, offset


class FrameConnectionCloseOk(Frame):
    name = 'connection.close_ok'
    method_id = METHOD_CONNECTION_CLOSE_OK

def decode_connection_close_ok(data, offset):
    frame = FrameConnectionCloseOk()
    return frame, offset


class FrameChannelOpenOk(Frame):
    name = 'channel.open_ok'
    method_id = METHOD_CHANNEL_OPEN_OK

def decode_channel_open_ok(data, offset):
    frame = FrameChannelOpenOk()
    (str_len,) = struct.unpack_from('!I', data, offset)
    offset += 4
    frame['channel_id'] = data[offset : offset+str_len]
    offset += str_len
    return frame, offset


class FrameChannelFlow(Frame):
    name = 'channel.flow'
    method_id = METHOD_CHANNEL_FLOW

def decode_channel_flow(data, offset):
    frame = FrameChannelFlow()
    (bits,) = struct.unpack_from('!B', data, offset)
    frame['active'] = bool(bits & 0x1)
    offset += 1
    return frame, offset


class FrameChannelFlowOk(Frame):
    name = 'channel.flow_ok'
    method_id = METHOD_CHANNEL_FLOW_OK

def decode_channel_flow_ok(data, offset):
    frame = FrameChannelFlowOk()
    (bits,) = struct.unpack_from('!B', data, offset)
    frame['active'] = bool(bits & 0x1)
    offset += 1
    return frame, offset


class FrameChannelClose(Frame):
    name = 'channel.close'
    method_id = METHOD_CHANNEL_CLOSE

def decode_channel_close(data, offset):
    frame = FrameChannelClose()
    (frame['reply_code'],
     str_len) = struct.unpack_from('!HB', data, offset)
    offset += 2+1
    frame['reply_text'] = data[offset : offset+str_len]
    offset += str_len
    (frame['class_id'],
     frame['method_id']) = struct.unpack_from('!HH', data, offset)
    offset += 2+2
    return frame, offset


class FrameChannelCloseOk(Frame):
    name = 'channel.close_ok'
    method_id = METHOD_CHANNEL_CLOSE_OK

def decode_channel_close_ok(data, offset):
    frame = FrameChannelCloseOk()
    return frame, offset


class FrameExchangeDeclareOk(Frame):
    name = 'exchange.declare_ok'
    method_id = METHOD_EXCHANGE_DECLARE_OK

def decode_exchange_declare_ok(data, offset):
    frame = FrameExchangeDeclareOk()
    return frame, offset


class FrameExchangeDeleteOk(Frame):
    name = 'exchange.delete_ok'
    method_id = METHOD_EXCHANGE_DELETE_OK

def decode_exchange_delete_ok(data, offset):
    frame = FrameExchangeDeleteOk()
    return frame, offset


class FrameExchangeBindOk(Frame):
    name = 'exchange.bind_ok'
    method_id = METHOD_EXCHANGE_BIND_OK

def decode_exchange_bind_ok(data, offset):
    frame = FrameExchangeBindOk()
    return frame, offset


class FrameExchangeUnbindOk(Frame):
    name = 'exchange.unbind_ok'
    method_id = METHOD_EXCHANGE_UNBIND_OK

def decode_exchange_unbind_ok(data, offset):
    frame = FrameExchangeUnbindOk()
    return frame, offset


class FrameQueueDeclareOk(Frame):
    name = 'queue.declare_ok'
    method_id = METHOD_QUEUE_DECLARE_OK

def decode_queue_declare_ok(data, offset):
    frame = FrameQueueDeclareOk()
    (str_len,) = struct.unpack_from('!B', data, offset)
    offset += 1
    frame['queue'] = data[offset : offset+str_len]
    offset += str_len
    (frame['message_count'],
     frame['consumer_count']) = struct.unpack_from('!II', data, offset)
    offset += 4+4
    return frame, offset


class FrameQueueBindOk(Frame):
    name = 'queue.bind_ok'
    method_id = METHOD_QUEUE_BIND_OK

def decode_queue_bind_ok(data, offset):
    frame = FrameQueueBindOk()
    return frame, offset


class FrameQueuePurgeOk(Frame):
    name = 'queue.purge_ok'
    method_id = METHOD_QUEUE_PURGE_OK

def decode_queue_purge_ok(data, offset):
    frame = FrameQueuePurgeOk()
    (frame['message_count'],) = struct.unpack_from('!I', data, offset)
    offset += 4
    return frame, offset


class FrameQueueDeleteOk(Frame):
    name = 'queue.delete_ok'
    method_id = METHOD_QUEUE_DELETE_OK

def decode_queue_delete_ok(data, offset):
    frame = FrameQueueDeleteOk()
    (frame['message_count'],) = struct.unpack_from('!I', data, offset)
    offset += 4
    return frame, offset


class FrameQueueUnbindOk(Frame):
    name = 'queue.unbind_ok'
    method_id = METHOD_QUEUE_UNBIND_OK

def decode_queue_unbind_ok(data, offset):
    frame = FrameQueueUnbindOk()
    return frame, offset


class FrameBasicQosOk(Frame):
    name = 'basic.qos_ok'
    method_id = METHOD_BASIC_QOS_OK

def decode_basic_qos_ok(data, offset):
    frame = FrameBasicQosOk()
    return frame, offset


class FrameBasicConsumeOk(Frame):
    name = 'basic.consume_ok'
    method_id = METHOD_BASIC_CONSUME_OK

def decode_basic_consume_ok(data, offset):
    frame = FrameBasicConsumeOk()
    (str_len,) = struct.unpack_from('!B', data, offset)
    offset += 1
    frame['consumer_tag'] = data[offset : offset+str_len]
    offset += str_len
    return frame, offset


class FrameBasicCancelOk(Frame):
    name = 'basic.cancel_ok'
    method_id = METHOD_BASIC_CANCEL_OK

def decode_basic_cancel_ok(data, offset):
    frame = FrameBasicCancelOk()
    (str_len,) = struct.unpack_from('!B', data, offset)
    offset += 1
    frame['consumer_tag'] = data[offset : offset+str_len]
    offset += str_len
    return frame, offset


class FrameBasicReturn(Frame):
    name = 'basic.return'
    method_id = METHOD_BASIC_RETURN
    has_content = True
    class_id = CLASS_BASIC

def decode_basic_return(data, offset):
    frame = FrameBasicReturn()
    (frame['reply_code'],
     str_len) = struct.unpack_from('!HB', data, offset)
    offset += 2+1
    frame['reply_text'] = data[offset : offset+str_len]
    offset += str_len
    (str_len,) = struct.unpack_from('!B', data, offset)
    offset += 1
    frame['exchange'] = data[offset : offset+str_len]
    offset += str_len
    (str_len,) = struct.unpack_from('!B', data, offset)
    offset += 1
    frame['routing_key'] = data[offset : offset+str_len]
    offset += str_len
    return frame, offset


class FrameBasicDeliver(Frame):
    name = 'basic.deliver'
    method_id = METHOD_BASIC_DELIVER
    has_content = True
    class_id = CLASS_BASIC

def decode_basic_deliver(data, offset):
    frame = FrameBasicDeliver()
    (str_len,) = struct.unpack_from('!B', data, offset)
    offset += 1
    frame['consumer_tag'] = data[offset : offset+str_len]
    offset += str_len
    (frame['delivery_tag'],
     bits,
     str_len) = struct.unpack_from('!QBB', data, offset)
    frame['redelivered'] = bool(bits & 0x1)
    offset += 8+1+1
    frame['exchange'] = data[offset : offset+str_len]
    offset += str_len
    (str_len,) = struct.unpack_from('!B', data, offset)
    offset += 1
    frame['routing_key'] = data[offset : offset+str_len]
    offset += str_len
    return frame, offset


class FrameBasicGetOk(Frame):
    name = 'basic.get_ok'
    method_id = METHOD_BASIC_GET_OK
    has_content = True
    class_id = CLASS_BASIC

def decode_basic_get_ok(data, offset):
    frame = FrameBasicGetOk()
    (frame['delivery_tag'],
     bits,
     str_len) = struct.unpack_from('!QBB', data, offset)
    frame['redelivered'] = bool(bits & 0x1)
    offset += 8+1+1
    frame['exchange'] = data[offset : offset+str_len]
    offset += str_len
    (str_len,) = struct.unpack_from('!B', data, offset)
    offset += 1
    frame['routing_key'] = data[offset : offset+str_len]
    offset += str_len
    (frame['message_count'],) = struct.unpack_from('!I', data, offset)
    offset += 4
    return frame, offset


class FrameBasicGetEmpty(Frame):
    name = 'basic.get_empty'
    method_id = METHOD_BASIC_GET_EMPTY

def decode_basic_get_empty(data, offset):
    frame = FrameBasicGetEmpty()
    (str_len,) = struct.unpack_from('!B', data, offset)
    offset += 1
    frame['cluster_id'] = data[offset : offset+str_len]
    offset += str_len
    return frame, offset


class FrameBasicAck(Frame):
    name = 'basic.ack'
    method_id = METHOD_BASIC_ACK

def decode_basic_ack(data, offset):
    frame = FrameBasicAck()
    (frame['delivery_tag'],
     bits) = struct.unpack_from('!QB', data, offset)
    frame['multiple'] = bool(bits & 0x1)
    offset += 8+1
    return frame, offset


class FrameBasicRecoverOk(Frame):
    name = 'basic.recover_ok'
    method_id = METHOD_BASIC_RECOVER_OK

def decode_basic_recover_ok(data, offset):
    frame = FrameBasicRecoverOk()
    return frame, offset


class FrameConfirmSelectOk(Frame):
    name = 'confirm.select_ok'
    method_id = METHOD_CONFIRM_SELECT_OK

def decode_confirm_select_ok(data, offset):
    frame = FrameConfirmSelectOk()
    return frame, offset



METHODS = {
    METHOD_CONNECTION_START:        decode_connection_start,
    METHOD_CONNECTION_SECURE:       decode_connection_secure,
    METHOD_CONNECTION_TUNE:         decode_connection_tune,
    METHOD_CONNECTION_OPEN_OK:      decode_connection_open_ok,
    METHOD_CONNECTION_CLOSE:        decode_connection_close,
    METHOD_CONNECTION_CLOSE_OK:     decode_connection_close_ok,
    METHOD_CHANNEL_OPEN_OK:         decode_channel_open_ok,
    METHOD_CHANNEL_FLOW:            decode_channel_flow,
    METHOD_CHANNEL_FLOW_OK:         decode_channel_flow_ok,
    METHOD_CHANNEL_CLOSE:           decode_channel_close,
    METHOD_CHANNEL_CLOSE_OK:        decode_channel_close_ok,
    METHOD_EXCHANGE_DECLARE_OK:     decode_exchange_declare_ok,
    METHOD_EXCHANGE_DELETE_OK:      decode_exchange_delete_ok,
    METHOD_EXCHANGE_BIND_OK:        decode_exchange_bind_ok,
    METHOD_EXCHANGE_UNBIND_OK:      decode_exchange_unbind_ok,
    METHOD_QUEUE_DECLARE_OK:        decode_queue_declare_ok,
    METHOD_QUEUE_BIND_OK:           decode_queue_bind_ok,
    METHOD_QUEUE_PURGE_OK:          decode_queue_purge_ok,
    METHOD_QUEUE_DELETE_OK:         decode_queue_delete_ok,
    METHOD_QUEUE_UNBIND_OK:         decode_queue_unbind_ok,
    METHOD_BASIC_QOS_OK:            decode_basic_qos_ok,
    METHOD_BASIC_CONSUME_OK:        decode_basic_consume_ok,
    METHOD_BASIC_CANCEL_OK:         decode_basic_cancel_ok,
    METHOD_BASIC_RETURN:            decode_basic_return,
    METHOD_BASIC_DELIVER:           decode_basic_deliver,
    METHOD_BASIC_GET_OK:            decode_basic_get_ok,
    METHOD_BASIC_GET_EMPTY:         decode_basic_get_empty,
    METHOD_BASIC_ACK:               decode_basic_ack,
    METHOD_BASIC_RECOVER_OK:        decode_basic_recover_ok,
    METHOD_CONFIRM_SELECT_OK:       decode_confirm_select_ok,
}


def decode_basic_properties(data, offset):
    props = {}
    flags, = struct.unpack_from('!H', data, offset)
    offset += 2
    assert (flags & 0x01) == 0
    if (flags & 0x8000): # 1 << 15
        (str_len,) = struct.unpack_from('!B', data, offset)
        offset += 1
        props['content_type'] = data[offset : offset+str_len]
        offset += str_len
    if (flags & 0x4000): # 1 << 14
        (str_len,) = struct.unpack_from('!B', data, offset)
        offset += 1
        props['content_encoding'] = data[offset : offset+str_len]
        offset += str_len
    if (flags & 0x2000): # 1 << 13
        props['headers'], offset = table.decode(data, offset)
    if (flags & 0x1000): # 1 << 12
        (props['delivery_mode'],) = struct.unpack_from('!B', data, offset)
        offset += 1
    if (flags & 0x0800): # 1 << 11
        (props['priority'],) = struct.unpack_from('!B', data, offset)
        offset += 1
    if (flags & 0x0400): # 1 << 10
        (str_len,) = struct.unpack_from('!B', data, offset)
        offset += 1
        props['correlation_id'] = data[offset : offset+str_len]
        offset += str_len
    if (flags & 0x0200): # 1 << 9
        (str_len,) = struct.unpack_from('!B', data, offset)
        offset += 1
        props['reply_to'] = data[offset : offset+str_len]
        offset += str_len
    if (flags & 0x0100): # 1 << 8
        (str_len,) = struct.unpack_from('!B', data, offset)
        offset += 1
        props['expiration'] = data[offset : offset+str_len]
        offset += str_len
    if (flags & 0x0080): # 1 << 7
        (str_len,) = struct.unpack_from('!B', data, offset)
        offset += 1
        props['message_id'] = data[offset : offset+str_len]
        offset += str_len
    if (flags & 0x0040): # 1 << 6
        (props['timestamp'],) = struct.unpack_from('!Q', data, offset)
        offset += 8
    if (flags & 0x0020): # 1 << 5
        (str_len,) = struct.unpack_from('!B', data, offset)
        offset += 1
        props['type_'] = data[offset : offset+str_len]
        offset += str_len
    if (flags & 0x0010): # 1 << 4
        (str_len,) = struct.unpack_from('!B', data, offset)
        offset += 1
        props['user_id'] = data[offset : offset+str_len]
        offset += str_len
    if (flags & 0x0008): # 1 << 3
        (str_len,) = struct.unpack_from('!B', data, offset)
        offset += 1
        props['app_id'] = data[offset : offset+str_len]
        offset += str_len
    if (flags & 0x0004): # 1 << 2
        (str_len,) = struct.unpack_from('!B', data, offset)
        offset += 1
        props['cluster_id'] = data[offset : offset+str_len]
        offset += str_len
    return props, offset


PROPS = {
    CLASS_BASIC: decode_basic_properties, 	# 60
}


# client_properties=None mechanism='PLAIN' response=None locale='en_US'
def encode_connection_start_ok(client_properties, mechanism, response, locale):
    client_properties_raw = table.encode(client_properties)
    return ( (0x01,
              ''.join((
                struct.pack('!I', METHOD_CONNECTION_START_OK),
                client_properties_raw,
                struct.pack('!B', len(mechanism)),
                mechanism,
                struct.pack('!I', len(response)),
                response,
                struct.pack('!B', len(locale)),
                locale,
              ))
           ), )

# response=None
def encode_connection_secure_ok(response):
    return ( (0x01,
              ''.join((
                struct.pack('!II', METHOD_CONNECTION_SECURE_OK, len(response)),
                response,
              ))
           ), )

# channel_max=0 frame_max=0 heartbeat=0
def encode_connection_tune_ok(channel_max, frame_max, heartbeat):
    return ( (0x01,
                struct.pack('!IHIH', METHOD_CONNECTION_TUNE_OK, channel_max, frame_max, heartbeat),
           ), )

# virtual_host='/' capabilities='' insist=False
def encode_connection_open(virtual_host):
    return ( (0x01,
              ''.join((
                struct.pack('!IB', METHOD_CONNECTION_OPEN, len(virtual_host)),
                virtual_host,
                "\x00\x00",
              ))
           ), )

# reply_code=None reply_text='' class_id=None method_id=None
def encode_connection_close(reply_code, reply_text, class_id, method_id):
    return ( (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_CONNECTION_CLOSE, reply_code, len(reply_text)),
                reply_text,
                struct.pack('!HH', class_id, method_id),
              ))
           ), )

# 
def encode_connection_close_ok():
    return ( (0x01,
                struct.pack('!I', METHOD_CONNECTION_CLOSE_OK),
           ), )

# out_of_band=''
def encode_channel_open(out_of_band):
    return ( (0x01,
                struct.pack('!IB', METHOD_CHANNEL_OPEN, 0),
           ), )

# active=None
def encode_channel_flow(active):
    return ( (0x01,
                struct.pack('!IB', METHOD_CHANNEL_FLOW, (active and 0x1 or 0)),
           ), )

# active=None
def encode_channel_flow_ok(active):
    return ( (0x01,
                struct.pack('!IB', METHOD_CHANNEL_FLOW_OK, (active and 0x1 or 0)),
           ), )

# reply_code=None reply_text='' class_id=None method_id=None
def encode_channel_close(reply_code, reply_text, class_id, method_id):
    return ( (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_CHANNEL_CLOSE, reply_code, len(reply_text)),
                reply_text,
                struct.pack('!HH', class_id, method_id),
              ))
           ), )

# 
def encode_channel_close_ok():
    return ( (0x01,
                struct.pack('!I', METHOD_CHANNEL_CLOSE_OK),
           ), )

# ticket=0 exchange=None type_='direct' passive=False durable=False auto_delete=False internal=False nowait=False arguments={}
def encode_exchange_declare(exchange, type_, passive, durable, auto_delete, internal, arguments):
    arguments_raw = table.encode(arguments)
    return ( (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_EXCHANGE_DECLARE, 0, len(exchange)),
                exchange,
                struct.pack('!B', len(type_)),
                type_,
                struct.pack('!B', (passive and 0x1 or 0) | (durable and 0x2 or 0) | (auto_delete and 0x4 or 0) | (internal and 0x8 or 0)),
                arguments_raw,
              ))
           ), )

# ticket=0 exchange=None if_unused=False nowait=False
def encode_exchange_delete(exchange, if_unused):
    return ( (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_EXCHANGE_DELETE, 0, len(exchange)),
                exchange,
                struct.pack('!B', (if_unused and 0x1 or 0)),
              ))
           ), )

# ticket=0 destination=None source=None routing_key='' nowait=False arguments={}
def encode_exchange_bind(destination, source, routing_key, arguments):
    arguments_raw = table.encode(arguments)
    return ( (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_EXCHANGE_BIND, 0, len(destination)),
                destination,
                struct.pack('!B', len(source)),
                source,
                struct.pack('!B', len(routing_key)),
                routing_key,
                "\x00",
                arguments_raw,
              ))
           ), )

# ticket=0 destination=None source=None routing_key='' nowait=False arguments={}
def encode_exchange_unbind(destination, source, routing_key, arguments):
    arguments_raw = table.encode(arguments)
    return ( (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_EXCHANGE_UNBIND, 0, len(destination)),
                destination,
                struct.pack('!B', len(source)),
                source,
                struct.pack('!B', len(routing_key)),
                routing_key,
                "\x00",
                arguments_raw,
              ))
           ), )

# ticket=0 queue='' passive=False durable=False exclusive=False auto_delete=False nowait=False arguments={}
def encode_queue_declare(queue, passive, durable, exclusive, auto_delete, arguments):
    arguments_raw = table.encode(arguments)
    return ( (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_QUEUE_DECLARE, 0, len(queue)),
                queue,
                struct.pack('!B', (passive and 0x1 or 0) | (durable and 0x2 or 0) | (exclusive and 0x4 or 0) | (auto_delete and 0x8 or 0)),
                arguments_raw,
              ))
           ), )

# ticket=0 queue='' exchange=None routing_key='' nowait=False arguments={}
def encode_queue_bind(queue, exchange, routing_key, arguments):
    arguments_raw = table.encode(arguments)
    return ( (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_QUEUE_BIND, 0, len(queue)),
                queue,
                struct.pack('!B', len(exchange)),
                exchange,
                struct.pack('!B', len(routing_key)),
                routing_key,
                "\x00",
                arguments_raw,
              ))
           ), )

# ticket=0 queue='' nowait=False
def encode_queue_purge(queue):
    return ( (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_QUEUE_PURGE, 0, len(queue)),
                queue,
                "\x00",
              ))
           ), )

# ticket=0 queue='' if_unused=False if_empty=False nowait=False
def encode_queue_delete(queue, if_unused, if_empty):
    return ( (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_QUEUE_DELETE, 0, len(queue)),
                queue,
                struct.pack('!B', (if_unused and 0x1 or 0) | (if_empty and 0x2 or 0)),
              ))
           ), )

# ticket=0 queue='' exchange=None routing_key='' arguments={}
def encode_queue_unbind(queue, exchange, routing_key, arguments):
    arguments_raw = table.encode(arguments)
    return ( (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_QUEUE_UNBIND, 0, len(queue)),
                queue,
                struct.pack('!B', len(exchange)),
                exchange,
                struct.pack('!B', len(routing_key)),
                routing_key,
                arguments_raw,
              ))
           ), )

# prefetch_size=0 prefetch_count=0 global_=False
def encode_basic_qos(prefetch_size, prefetch_count, global_):
    return ( (0x01,
                struct.pack('!IIHB', METHOD_BASIC_QOS, prefetch_size, prefetch_count, (global_ and 0x1 or 0)),
           ), )

# ticket=0 queue='' consumer_tag='' no_local=False no_ack=False exclusive=False nowait=False arguments={}
def encode_basic_consume(queue, consumer_tag, no_local, no_ack, exclusive, arguments):
    arguments_raw = table.encode(arguments)
    return ( (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_BASIC_CONSUME, 0, len(queue)),
                queue,
                struct.pack('!B', len(consumer_tag)),
                consumer_tag,
                struct.pack('!B', (no_local and 0x1 or 0) | (no_ack and 0x2 or 0) | (exclusive and 0x4 or 0)),
                arguments_raw,
              ))
           ), )

# consumer_tag=None nowait=False
def encode_basic_cancel(consumer_tag):
    return ( (0x01,
              ''.join((
                struct.pack('!IB', METHOD_BASIC_CANCEL, len(consumer_tag)),
                consumer_tag,
                "\x00",
              ))
           ), )

# ticket=0 exchange='' routing_key='' mandatory=False immediate=False user_headers={} payload='' frame_size=None
def encode_basic_publish(exchange, routing_key, mandatory, immediate, user_headers, body, frame_size):
    props, headers = split_headers(user_headers, BASIC_PROPS_SET)
    if headers:
        props['headers'] = headers
    return [ (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_BASIC_PUBLISH, 0, len(exchange)),
                exchange,
                struct.pack('!B', len(routing_key)),
                routing_key,
                struct.pack('!B', (mandatory and 0x1 or 0) | (immediate and 0x2 or 0)),
              ))
           ),
           encode_basic_properties(len(body), props),
        ] + encode_body(body, frame_size)

# ticket=0 queue='' no_ack=False
def encode_basic_get(queue, no_ack):
    return ( (0x01,
              ''.join((
                struct.pack('!IHB', METHOD_BASIC_GET, 0, len(queue)),
                queue,
                struct.pack('!B', (no_ack and 0x1 or 0)),
              ))
           ), )

# delivery_tag=0 multiple=False
def encode_basic_ack(delivery_tag, multiple):
    return ( (0x01,
                struct.pack('!IQB', METHOD_BASIC_ACK, delivery_tag, (multiple and 0x1 or 0)),
           ), )

# delivery_tag=None requeue=True
def encode_basic_reject(delivery_tag, requeue):
    return ( (0x01,
                struct.pack('!IQB', METHOD_BASIC_REJECT, delivery_tag, (requeue and 0x1 or 0)),
           ), )

# requeue=False
def encode_basic_recover_async(requeue):
    return ( (0x01,
                struct.pack('!IB', METHOD_BASIC_RECOVER_ASYNC, (requeue and 0x1 or 0)),
           ), )

# requeue=False
def encode_basic_recover(requeue):
    return ( (0x01,
                struct.pack('!IB', METHOD_BASIC_RECOVER, (requeue and 0x1 or 0)),
           ), )

# delivery_tag=0 multiple=False requeue=True
def encode_basic_nack(delivery_tag, multiple, requeue):
    return ( (0x01,
                struct.pack('!IQB', METHOD_BASIC_NACK, delivery_tag, (multiple and 0x1 or 0) | (requeue and 0x2 or 0)),
           ), )

# nowait=False
def encode_confirm_select():
    return ( (0x01,
                struct.pack('!IB', METHOD_CONFIRM_SELECT, 0),
           ), )

BASIC_PROPS_SET = set((
    "content_type",      # shortstr
    "content_encoding",  # shortstr
    "headers",           # table
    "delivery_mode",     # octet
    "priority",          # octet
    "correlation_id",    # shortstr
    "reply_to",          # shortstr
    "expiration",        # shortstr
    "message_id",        # shortstr
    "timestamp",         # timestamp
    "type_",             # shortstr
    "user_id",           # shortstr
    "app_id",            # shortstr
    "cluster_id",        # shortstr
    ))

ENCODE_PROPS_BASIC = {
    'content_type': (
        0,
        0x8000, # (1 << 15)
        lambda val: ''.join((
                struct.pack('!B', len(val)),
                val,
        )) ),
    'content_encoding': (
        1,
        0x4000, # (1 << 14)
        lambda val: ''.join((
                struct.pack('!B', len(val)),
                val,
        )) ),
    'headers': (
        2,
        0x2000, # (1 << 13)
        lambda val: table.encode(val)
        ),
    'delivery_mode': (
        3,
        0x1000, # (1 << 12)
        lambda val: struct.pack('!B', val)
        ),
    'priority': (
        4,
        0x0800, # (1 << 11)
        lambda val: struct.pack('!B', val)
        ),
    'correlation_id': (
        5,
        0x0400, # (1 << 10)
        lambda val: ''.join((
                struct.pack('!B', len(val)),
                val,
        )) ),
    'reply_to': (
        6,
        0x0200, # (1 << 9)
        lambda val: ''.join((
                struct.pack('!B', len(val)),
                val,
        )) ),
    'expiration': (
        7,
        0x0100, # (1 << 8)
        lambda val: ''.join((
                struct.pack('!B', len(val)),
                val,
        )) ),
    'message_id': (
        8,
        0x0080, # (1 << 7)
        lambda val: ''.join((
                struct.pack('!B', len(val)),
                val,
        )) ),
    'timestamp': (
        9,
        0x0040, # (1 << 6)
        lambda val: struct.pack('!Q', val)
        ),
    'type_': (
        10,
        0x0020, # (1 << 5)
        lambda val: ''.join((
                struct.pack('!B', len(val)),
                val,
        )) ),
    'user_id': (
        11,
        0x0010, # (1 << 4)
        lambda val: ''.join((
                struct.pack('!B', len(val)),
                val,
        )) ),
    'app_id': (
        12,
        0x0008, # (1 << 3)
        lambda val: ''.join((
                struct.pack('!B', len(val)),
                val,
        )) ),
    'cluster_id': (
        13,
        0x0004, # (1 << 2)
        lambda val: ''.join((
                struct.pack('!B', len(val)),
                val,
        )) ),
}

def encode_basic_properties(body_size, props):
    pieces = ['']*14
    flags = 0
    enc = ENCODE_PROPS_BASIC

    for key in BASIC_PROPS_SET & set(props.iterkeys()):
        i, f, fun = enc[key]
        flags |= f
        pieces[i] = fun(props[key])

    return (0x02, ''.join((
        struct.pack('!HHQH',
                    CLASS_BASIC, 0, body_size, flags),
        ''.join(pieces),
        ))
        )


def split_headers(user_headers, properties_set):
    props = {}
    headers = {}
    for key, value in user_headers.iteritems():
        if key in properties_set:
            props[key] = value
        else:
            headers[key] = value
    return props, headers

def encode_body(body, frame_size):
    limit = frame_size - 7 - 1   # spec is broken...
    r = []
    while body:
        payload, body = body[:limit], body[limit:]
        r.append( (0x03, payload) )
    return r




import os
import sys
import setuptools

# Some filesystems don't support hard links. Use the power of
# monkeypatching to overcome the problem.
import os, shutil
os.link = shutil.copy


if not os.path.exists("puka/spec.py"):
    print >> sys.stderr, "Run 'make' first."
    sys.exit(1)


setuptools.setup(name='puka',
      version=file('VERSION').read().strip(),
      description='Puka - the opinionated RabbitMQ client',
      author='Marek Majkowski',
      author_email='marek@rabbitmq.com',
      url='http://github.com/majek/puka#readme',
      packages=['puka'],
      platforms=['any'],
      license='MIT',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
      zip_safe = True,
      )



# Autogenerated - do not edit

class AMQPError(Exception): pass
class AMQPSoftError(AMQPError): pass
class AMQPHardError(AMQPError): pass

class ContentTooLarge(AMQPSoftError):
    reply_code = 311
class NoRoute(AMQPSoftError):
    reply_code = 312
class NoConsumers(AMQPSoftError):
    reply_code = 313
class AccessRefused(AMQPSoftError):
    reply_code = 403
class NotFound(AMQPSoftError):
    reply_code = 404
class ResourceLocked(AMQPSoftError):
    reply_code = 405
class PreconditionFailed(AMQPSoftError):
    reply_code = 406
class ConnectionForced(AMQPHardError):
    reply_code = 320
class InvalidPath(AMQPHardError):
    reply_code = 402
class FrameError(AMQPHardError):
    reply_code = 501
class AMQPSyntaxError(AMQPHardError):
    reply_code = 502
class CommandInvalid(AMQPHardError):
    reply_code = 503
class ChannelError(AMQPHardError):
    reply_code = 504
class UnexpectedFrame(AMQPHardError):
    reply_code = 505
class ResourceError(AMQPHardError):
    reply_code = 506
class NotAllowed(AMQPHardError):
    reply_code = 530
class AMQPNotImplemented(AMQPHardError):
    reply_code = 540
class InternalError(AMQPHardError):
    reply_code = 541

ERRORS = {
    311: ContentTooLarge,
    312: NoRoute,
    313: NoConsumers,
    403: AccessRefused,
    404: NotFound,
    405: ResourceLocked,
    406: PreconditionFailed,
    320: ConnectionForced,
    402: InvalidPath,
    501: FrameError,
    502: AMQPSyntaxError,
    503: CommandInvalid,
    504: ChannelError,
    505: UnexpectedFrame,
    506: ResourceError,
    530: NotAllowed,
    540: AMQPNotImplemented,
    541: InternalError,
}

